<?php
/**
 * WP CLI commands for DMG Media
 *
 * @package DMG_WP_CLI
 */

/**
 * Plugin Name: DMG CLI
 * Plugin URI:  https://github.com/ANMedia/dmg-cli-wp-plugin
 * Description: WP CLI commands for DMG Media
 * Version:     1.1.0
 * Author:      markparolisi
 */

if( !class_exists( 'WPCOM_VIP_CLI_Command' ) ) {
    return;
}

/**
 * CLI commands for DMG Media
 */
class DMG_WP_CLI extends WPCOM_VIP_CLI_Command {


    /**
     * Helper method to output a dry-run message
     *
     * @param boolean $dry_run Whether or not doing a dry run
     * @return void
     */
    private static function dry_run_message( $dry_run = false ) {
        $dry_run_message = ( $dry_run ) ? 'Running in dry-run mode.' : 'Running in production mode!';
		WP_CLI::line( $dry_run_message );
    }

    /**
	 * Utility to remove a specific capability
	 *
	 * ## OPTIONS
	 *
	 * cap: the capability name
	 *
	 * ## EXAMPLES
	 *
	 * wp dmg user-remove-cap --cap="cap_name" --dry-run=true
	 *
	 * @subcommand  user-remove-cap
	 * @synopsis [--cap] [--dry-run=<bool>]
	 */
	public function user_remove_cap( $args, $assoc_args ) {

		$defaults = [
			'dry-run' => true,
			'cap'     => '',
		];

		$assoc_args = array_merge( $defaults, $assoc_args );

		$dry_run = ( isset( $assoc_args['dry-run'] ) ) ? filter_var( $assoc_args['dry-run'], FILTER_VALIDATE_BOOLEAN ) : false;
        self::dry_run_message($dry_run);

		if ( $assoc_args['cap'] !== '' ) {
			global $wp_roles;
			foreach ( array_keys( $wp_roles->roles ) as $role ) {
				WP_CLI::line( sprintf( 'Removing "%s" from "%s"', $assoc_args['cap'], $role ) );
				if ( ! $dry_run ) {
					$wp_roles->remove_cap( $role, $assoc_args['cap'] );
				}
			}
            WP_CLI::success('User Caps removed.');
		} else {
			WP_CLI::error( 'Please specify the capability to remove. Usage: wp metro remove-cap --cap=<cap>' );
		}
	}


	/**
	 * Set the role to subscriber if the user role is missing
	 *
	 * @subcommand user-fill-missing-role
	 *
	 * @synopsis [--dry-run]
	 *
	 * wp dmg user-fill-missing-role --dry-run=true
	 */
	public function user_fill_missing_role( $args, $assoc_args ) {

		$count = 0;

		$defaults = [
			'dry-run' => true,
		];

		$assoc_args = array_merge( $defaults, $assoc_args );

		$dry_run = isset( $assoc_args['dry-run'] ) ? filter_var( $assoc_args['dry-run'], FILTER_VALIDATE_BOOLEAN ) : true;

        self::dry_run_message($dry_run);

		$query_args = [ 'orderby' => 'user_registered' ];
		$user_query = new \WP_User_Query( $query_args );

		if ( ! empty( $user_query->get_results() ) ) {
			$this->start_bulk_operation();

			foreach ( $user_query->get_results() as $user ) {

				if ( empty ( $user->roles ) ) {
					WP_CLI::log( "Missing role for user $user->ID - $user->user_login." );
					$count ++;
					if ( ! $dry_run ) {
						WP_CLI::runcommand( "user set-role $user->ID subscriber", [ 'exit_error' => false ] );
						sleep( 1 );
					}
				}
			}

			if ( ! $dry_run ) {
				WP_CLI::success( sprintf( '%d users successfully updated.', $count ) );
			} else {
				WP_CLI::success( sprintf( '%d users will be updated.', $count ) );
			}

			$this->end_bulk_operation();
		} else {
			WP_CLI::log( 'No users found' );
		}
	}

	/**
	 * Append a taxonomy to posts that meet separate taxonomy criteria
	 *
	 * ## OPTIONS
	 *
	 * <taxonomy>
	 * : The taxonomy of the term to append
	 *
	 * <term>
	 * : The term to append
	 *
	 * [--term-field=<term-field>]
	 * : The field the term is expressed as; default slug
	 *
	 * [--date-after=<date-after>]
	 * : Start date for querying posts
	 *
	 * [--date-before=<date-before>]
	 * : End date for querying posts
	 *
	 * [--<field>=<value>]
	 * : Args for searching existing posts
	 *
	 *   There are three possible parameters that can be passed:
	 *
	 *      1) --search-<taxonomy>=<term>: The taxonomy term to query by
	 *      2) --search-<taxonomy>-field=<field>: The field to query the search taxonomy term by; default slug
	 *      3) --search-<taxonomy>-include-children=<include>: Whether to include children in taxonomy term search;
	 *
	 * [--search-relationship=<search-relationship>]
	 * : Relationship of tax query; default OR
	 *
	 * [--search-string=<search-string>]
	 * : The text string to search post content for
	 *
	 * [--search-post-type=<post-type>]
	 * : The post type to search; default post
	 *
	 * [--case-sensitive=<case-sensitive>]
	 * : If search-string is passed, enforce case sensitivity
	 *
	 * [--ignore-shortcodes=<ignore-shortcodes>]
	 * : If search-string is passed, ignore shortcodes
	 *
	 * [--append=<append>]
	 * : If true, append terms; if false, replace existing terms; default true
	 *
	 * [--dry-run=<dry-run>]
	 * : Perform the command without updating the database
	 *
	 * ## EXAMPLES
	 *
	 * wp dmg auto-term post_tag pizza --search-post_tag=cheese --dry-run=true
	 *
	 * @subcommand auto-term
	 */
	public function auto_term( $args, $assoc_args ) {

		// Is this a dry run?
		$dry_run = ( isset( $assoc_args['dry-run'] ) ) ? filter_var( $assoc_args['dry-run'], FILTER_VALIDATE_BOOLEAN ) : false;
		self::dry_run_message( $dry_run );

		// Make sure the taxonomy is valid
		if ( empty( $args[0] ) ) {
			WP_CLI::error( 'No taxonomy specified.' );
		}

		$taxonomy = get_taxonomy( $args[0] );
		if ( ! is_a( $taxonomy, 'WP_Taxonomy' ) ) {
			WP_CLI::error( 'Invalid taxonomy.' );
		}

		// Make sure the term is valid
		if ( empty( $args[1] ) ) {
			WP_CLI::error( 'No term specified.' );
		}

		// Determine the term field
		$term_field = 'slug';
		if ( isset( $assoc_args['term-field'] ) ) {
			if ( in_array( $assoc_args['term-field'], array( 'term_id', 'name', 'slug ') ) ) {
				$term_field = $assoc_args['term-field'];
			} else {
				WP_CLI::error( 'Invalid term field passed: ' . $assoc_args['term-field'] );
			}
		}

		// Retrieve the term object
		$term = get_term_by( $term_field, $args[1], $taxonomy->name );
		if ( ! is_a( $term, 'WP_Term' ) ) {
			WP_CLI::error( 'Term with ' . $term_field . ' ' .  $args[1] . ' in taxonomy ' . $taxonomy->name . ' not found.' );
		}

		// Defaults for other arguments
		$case_sensitive    = ( isset( $assoc_args['case-sensitive'] ) ) ? filter_var( $assoc_args['case-sensitive'], FILTER_VALIDATE_BOOLEAN ) : true;
		$ignore_shortcodes = ( isset( $assoc_args['ignore-shortcodes'] ) ) ? filter_var( $assoc_args['ignore-shortcodes'], FILTER_VALIDATE_BOOLEAN ) : true;
		$append            = ( isset( $assoc_args['append'] ) ) ? filter_var( $assoc_args['append'], FILTER_VALIDATE_BOOLEAN ) : true;

		// Determine the post type
		$post_type = 'post';
		if ( isset( $assoc_args['search-post-type'] ) ) {
			if ( post_type_exists( $assoc_args['search-post-type'] ) ) {
				$post_type = $assoc_args['search-post-type'];
			} else {
				WP_CLI::error( 'Invalid post type specified: ' . $assoc_args['search-post-type'] );
			}
		}

		// Baseline query args
		$query_args = array(
			'post_status' => 'publish',
			'post_type'   => $post_type,
		);

		$tax_queries = array();

		// Build a regex pattern
		$taxonomy_pattern = implode( '|', get_taxonomies() );

		// Look for search parameters
		foreach ( $assoc_args as $key => $value ) {

			// Match against the taxonomy term
			if ( ! preg_match( "/search-({$taxonomy_pattern})$/", $key, $matches ) ) {
				continue;
			}

			if ( empty( $matches[1] ) ) {
				continue;
			}

			$tax_query_taxonomy = $matches[1];

			// Determine the field to query by
			$tax_query_field = 'slug';
			if ( isset( $assoc_args["{$key}-field"] ) ) {
				if ( in_array( $assoc_args["{$key}-field"], array( 'term_id', 'name', 'slug' ), true ) ) {
					$tax_query_field = $assoc_args["{$key}-field"];
				} else {
					WP_CLI::error( "Invalid value passed for {$key}-field: " . $assoc_args["{$key}-field"] );
				}
			}

			// Do we want to include children for hierarchical terms?
			$tax_query_include_children = false;
			if ( isset( $assoc_args["{$key}-include-children"] ) ) {
				$tax_query_include_children = filter_var( $assoc_args["{$key}-include-children"], FILTER_VALIDATE_BOOLEAN );
			}

			// Add the tax query
			$tax_queries[] = array(
				'taxonomy'         => $tax_query_taxonomy,
				'field'            => $tax_query_field,
				'terms'            => $value,
				'include_children' => $tax_query_include_children,
			);
		}

		// Build tax_query array
		if ( ! empty( $tax_queries ) ) {

			$query_args['tax_query'] = array();

			// Specify search relationship if we have more than one tax query
			if ( count( $tax_queries ) > 1 ) {

				// Default to OR
				$search_relationship = 'OR';

				if ( isset( $assoc_args['search-relationship'] ) ) {
					if ( in_array( strtoupper( $assoc_args['search-relationship'] ), array( 'AND', 'OR' ), true ) ) {
						$search_relationship = strtoupper( $assoc_args['search-relationship'] );
					} else {
						WP_CLI::error( "Invalid value passed for search-relationship: " . $assoc_args['search-relationship'] );
					}
				}

				$query_args['tax_query']['relation'] = $search_relationship;
			}

			$query_args['tax_query'] = array_merge( $query_args['tax_query'], $tax_queries );
		}

		// Get date query parameters from CLI args
		$date_query = self::get_date_query( $assoc_args );
		if ( ! empty( $date_query ) ) {
			$query_args['date_query'] = $date_query;
		}

		// Specify a search string
		if ( ! empty( $assoc_args['search-string'] ) ) {
			$query_args['s'] = $assoc_args['search-string'];
		}

		// Args to pass to the iterator
		$iterate_args = array_merge(
			$assoc_args,
			array(
				'term'              => $term,
				'dry_run'           => $dry_run,
				'case_sensitive'    => $case_sensitive,
				'ignore_shortcodes' => $ignore_shortcodes,
				'append'            => $append,
			)
		);

		// Iterate over posts with callback
		$this->iterate_posts( array( __CLASS__, 'auto_term_handler' ), $query_args, $iterate_args );
	}

	/**
	 * Callback for auto-term iterator
	 *
	 * @param WP_Post $post The post object
	 * @param array $args Callback arguments
	 */
	public static function auto_term_handler( $post, $args ) {

		WP_CLI::line( 'Checking post #' . $post->ID . ' -- ' . $post->post_title );

		// Make sure iterator has received a WP_Term object
		if ( empty( $args['term'] ) || ! is_a( $args['term'], 'WP_Term' ) ) {
			WP_CLI::error( 'The iterator needs a WP_Term object.' );
		}

		// Do a secondary case sensitive text check for search terms
		if ( ! empty( $args['search-string'] ) ) {

			// Make sure search method obeys case sensitivity argument
			$search  = $args['search-string'];
			$content = $post->post_content;
			if ( $args['case_sensitive'] !== true ) {
				$search  = strtolower( $search );
				$content = strtolower( $content );
			}

			// If we are ignoring shortcodes, strip them out here
			if ( $args['ignore_shortcodes'] === true ) {
				$pattern = get_shortcode_regex();
				$content = preg_replace( '/' . $pattern . '/s', '', $content );
			}

			// Skip post due to case mismatch
			if ( strpos( $content, $search ) === false ) {
				WP_CLI::line( 'Post #' . $post->ID . ' returned in search results, but does not contain exact text match. Skipping...' );
				return;
			}
		}

		// Append the term
		if ( ! $args['dry_run'] ) {
			$result = wp_set_object_terms( $post->ID, $args['term']->term_id, $args['term']->taxonomy, $args['append'] );
			if ( is_array( $result ) ) {
				$success = true;
			}
		} else {
			$success = true;
		}

		if ( $success ) {

			if ( ! $args['dry_run'] ) {
				WP_CLI::line( 'Post #' . $post->ID . ' was updated with ' . $args['term']->taxonomy . ' ' . $args['term']->name . '.' );
			} else {
				WP_CLI::line( 'Post #' . $post->ID . ' will be updated with ' . $args['term']->taxonomy . ' ' . $args['term']->name . '.' );
			}
		} else {

			WP_CLI::line( 'Something went wrong appending terms to post #' . $post->ID );
		}
	}

	/**
	 * Helper to get date query param from array of CLI arguments
	 *
	 * This expects a "date-before" and/or a "date-after"
	 * array key to work properly
	 *
	 * @param array $args The CLI arguments
	 * @return array A date_query parameter ready for WP_Query
	 */
	public static function get_date_query( $args ) {

		// Add date query if set
		$date_query = [];

		// Set the date after parameter
		if ( isset( $args['date-after'] ) ) {
			$pieces = explode( '-', $args['date-after'] );
			if ( isset( $pieces[0] )
			     && isset( $pieces[1] )
			     && isset( $pieces[2] )
			     && wp_checkdate( (int) $pieces[1], (int) $pieces[2], (int) $pieces[0], $args['date-after'] )
			) {
				$date_query['after'] = [
					'year'  => (int) $pieces[0],
					'month' => (int) $pieces[1],
					'day'   => (int) $pieces[2],
				];
			} else {
				WP_CLI::error( 'Invalid "date after" supplied.' );
			}
		}

		// End date
		if ( isset( $args['date-before'] ) ) {
			$pieces = explode( '-', $args['date-before'] );
			if ( isset( $pieces[0] )
			     && isset( $pieces[1] )
			     && isset( $pieces[2] )
			     && wp_checkdate( (int) $pieces[1], (int) $pieces[2], (int) $pieces[0], $args['date-before'] )
			) {
				$date_query['before'] = [
					'year'  => (int) $pieces[0],
					'month' => (int) $pieces[1],
					'day'   => (int) $pieces[2],
				];
			} else {
				WP_CLI::error( 'Invalid "date before" supplied.' );
			}
		}

		return $date_query;
	}

	/**
	 * Wrapper around wp_query that performs pagination and cleanup. Correctly iterates over all
	 * posts, even if posts self-modify to be removed from the query.
	 *
	 * @param string function( \WP_Post $post ) callback that operates on a WP_Post
	 * @param array  WP_Query arguments to select posts
	 * @param array  Other arguments to pass to callback
	 * @param string Order of execution. 'ASC' for ascending, 'DESC' for descending.
	 */
	protected function iterate_posts( $callback, $query_args, $func_args = [], $order = 'ASC' ) {

		$posts_per_page = 1; // Speed up first query to count total results
		$page           = 1;
		// This function tries to avoid the problem of missing posts if the CLI function performs an action that removes a post
		// from the query. It does it by running on the last object in the query and going in reverse order, effectively reversing
		// the ASC<->DESC order. We reverse the apparent order to the caller so that appears sensible still makes sense.
		if ( strcasecmp( $order, 'ASC' ) === 0 ) {
			$order = 'DESC';
		} else {
			$order = 'ASC';
		}

		// Pagination fails with 'p' query - change to use post_in if exists
		if ( isset( $query_args['p'] ) ) {
			if ( isset( $query_args['post__in'] ) ) {
				WP_CLI::warning( "'post__in' and 'p' query set - ignore 'post__in'" );
			}
			$query_args['post__in'] = [ $query_args['p'] ];
			unset( $query_args['p'] );
		}

		if ( ! isset( $query_args['posts_per_page'] ) ) {
			$query_args['posts_per_page'] = $posts_per_page;
		}

		if ( ! isset( $query_args['paged'] ) ) {
			$query_args['paged'] = $page;
		}

		if ( ! isset( $query_args['order'] ) ) {
			$query_args['order'] = $order;
		}

		// Run the query twice so that we can iterate in reverse order
		$query                        = new \WP_Query( $query_args );
		$total                        = $query->found_posts;
		$posts_per_page               = 100;
		$query_args['paged']          = floor( $total / $posts_per_page ) + 1;
		$query_args['posts_per_page'] = $posts_per_page;

		do {
			$query = new \WP_Query( $query_args );

			WP_CLI::line( 'Query: ' . wp_json_encode( $query->query ) );
			WP_CLI::line( 'Found: ' . count( $query->posts ) );

			// Iterate through found posts backwards so that we don't
			// Run into pagination issues if our CLI command removes
			// posts from the resulting query
			if ( $query->have_posts() ) {
				for ( $i = $query->post_count - 1; $i >= 0; $i -- ) {
					call_user_func( $callback, $query->posts[ $i ], $func_args );
				}
			}

			// Increase the page count
			$query_args['paged'] --;

			// Free up memory
			$this->stop_the_insanity();
		} while ( $query_args['paged'] >= 1 );

	}

}

WP_CLI::add_command( 'dmg', 'DMG_WP_CLI' );
